// OneWireServlet.java - A demo 1-Wire servlet for TINI.
//
// Copyright (C) 1999-2002  Smart Software Consulting
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Smart Software Consulting
// 1688 Silverwood Court
// Danville, CA  94526-3079
// USA
//
// http://www.smartsc.com
//

import java.io.*;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.servlet.*;
import javax.servlet.http.*;

import com.dalsemi.onewire.OneWireAccessProvider;
import com.dalsemi.onewire.OneWireException;
import com.dalsemi.onewire.adapter.DSPortAdapter;
import com.dalsemi.onewire.container.MemoryBank;
import com.dalsemi.onewire.container.OTPMemoryBank;
import com.dalsemi.onewire.container.PagedMemoryBank;
import com.dalsemi.onewire.container.OneWireContainer;
import com.dalsemi.onewire.utils.Address;

import com.smartsc.onewire.PortAdapterEnumerator;
import com.smartsc.onewire.PortAdapterEnumerator.Callback;

public class OneWireServlet
extends HttpServlet
{
	public void doPost( HttpServletRequest req, HttpServletResponse res)
	throws ServletException, IOException
	{
		doGet( req, res);
	}

	public final void doGet( HttpServletRequest req, HttpServletResponse res)
	throws ServletException, IOException
	{
		String adapter = req.getParameter( ADAPTER);
		String port = req.getParameter( PORT);
		String device = req.getParameter( DEVICE);
		String bankStr = req.getParameter( BANK);

		int bank = -1;
		try { bank = Integer.parseInt( bankStr); }
		catch( NumberFormatException nfe) { }

		// Do HTTP header stuff
		res.setContentType( "text/html");

		// Get Writer
		PrintWriter pw = res.getWriter();

		// Start HTML
		pw.println( "<html>");
		pw.println( "<head><title>1-Wire Servlet</title></head>");
		pw.println( "<body>");

		if( adapter != null && port != null)
		{
			DSPortAdapter pa = null;
			try
			{
				pa = getAdapter( adapter, port);

				try { registerOneWireContainerClasses( pa); }
				catch( OneWireException ibe) { log( ibe.toString()); }

				if( device != null && bank != -1)
				{
				    sendBankDetail( pw, pa.getDeviceContainer( device), bank);
				}
				else if( device != null)
				{
				    sendDeviceDetail( pw, pa.getDeviceContainer( device));
				}
				else
				{
				    sendPortAdapterDetail( pw, req, pa);
				}
			}
			catch( Exception e)
			{
				sendAlert( pw, e);
				pw.println( "<P>Error: " + e.getMessage());
			}
		}
		else
		{
			sendPortAdapterList( pw, req);
		}

		// Finish HTML
		pw.println("</body></html>");
	}

	public static final void sendPortAdapterList(
		final PrintWriter pw, final HttpServletRequest req)
	{
		pw.println( "<H1>1-Wire Port and Port Adapter Information</H1>");

		pw.println( "This servlet is a 1-wire explorer.  It uses the ");
		pw.println( "1-Wire classes to determine all of the 1-wire port ");
		pw.println( "adapters connected to this system.");
		pw.println( "<P>");
		pw.println( "The list below shows the 1-wire port adapters ");
		pw.println( "connected to this system and the ports to which ");
		pw.println( "they are connected. Click on [Details] to find ");
		pw.println( "out more about a specific 1-wire network.");
		pw.println( "<UL>");

		int portAdapterCount = PortAdapterEnumerator.enumerate(
			new PortAdapterEnumerator.Callback()
			{
				public void forEachPortAdapter( DSPortAdapter pa, String p)
				{
						pw.print  ( "<LI>A <b>");
						pw.print  ( pa.getAdapterName());
						pw.println( "</b> port adapter ");
						pw.println( "is connected to port <b>");
						pw.print  ( p);
						pw.print  ( "</b>. [<A HREF=");
						pw.print  ( req.getServletPath());
						pw.print  ( "?");
						pw.print  (
							buildQueryString( pa.getAdapterName(), p, null) );
						pw.print  ( ">Details</A>]");
				}
			});

		pw.println( "</UL>");

		pw.print( "Detected ");
		pw.print( portAdapterCount);
		pw.print( " port adapter");
		if( portAdapterCount != 1) pw.print( "s");
		pw.println( ".");
	}

	public static final void sendPortAdapterDetail(
		PrintWriter pw, HttpServletRequest req, DSPortAdapter pa)
	throws OneWireException
	{
		pw.println( "<H1>1-Wire Information</H1>");

		pw.println( "This page displays information about a particular ");
		pw.println( "1-wire network connected to this system.  The ");
		pw.println( "information is determined by using the 1-Wire classes.");
		pw.println( "On any given system, a 1-wire network can be uniquely ");
		pw.println( "identified by specifying a port and a port adapter.");

		pw.println( "<H2>Port Information</H2>");

		String port = pa.getPortName();
		print( pw, "Name", port);
		print( pw, "Description", pa.getPortTypeDescription());

		pw.println( "<H2>Port Adapter Information</H2>");

		pw.println( "<H3>Logical Information</H3>");

		String adapter = pa.getAdapterName();
		print( pw, "Name", adapter);
		print( pw, "Class version", pa.getClassVersion());

		pw.println( "<H3>Physical Information</H3>");

		print( pw, "ID", pa.getAdapterAddress());
		print( pw, "Version", pa.getAdapterVersion());

		pw.println( "<H3>Power Information</H3>");

		print( pw, "Can deliver power", pa.canDeliverPower());
		print( pw, "Can deliver smart power", pa.canDeliverSmartPower());
		print( pw, "Can program", pa.canProgram());

		pw.println( "<H3>Speed Information</H3>");

		print( pw, "Can overdrive", pa.canOverdrive());
		print( pw, "Can hyperdrive", pa.canHyperdrive());

		pw.println( "<H2>Currently Detected 1-Wire Devices</H2>");
		pw.println( "If a device specific servlet is available for a ");
		pw.println( "detected device, the device's ROM ID ");
		pw.println( "will be a link to the device specific servlet.");
		pw.println( "<DL>");

		int count = 0;

        // Clear any previous search restrictions
        pa.setSearchAllDevices();
        pa.targetAllFamilies();
		Enumeration ibEnum = pa.getAllDeviceContainers();
		while( ibEnum.hasMoreElements())
		{
			OneWireContainer ib = (OneWireContainer)ibEnum.nextElement();

			pw.print( "<DT>");

			pw.print(
				getRomIdHtml(
					adapter, port, ib.getAddress() ) );

			pw.print( " &#150; ");
			pw.print( htmlEscape( ib.getName()));
			pw.print( " (");
			pw.print( htmlEscape( ib.getAlternateNames()));
			pw.println( ")");

			pw.print( "<DD>");
			pw.println( htmlEscape( ib.getDescription()));
			pw.println( "<P>");
			++count;
		}
		pw.println( "</DL>");

		pw.print( "Detected ");
		pw.print( count);
		pw.print( " 1-Wire device");
		if( count != 1) pw.print( "s");
		pw.print( ".");
	}

	public static final void sendDeviceDetail(
		PrintWriter pw, OneWireContainer owd)
	throws OneWireException
	{
		pw.print( "<H1>1-Wire Memory Bank Information for Device ");
		pw.print( owd.getAddressAsString());
		pw.println( "</H1>");

		pw.println( "This page displays memory bank information for a ");
		pw.println( "1-wire device connected to this system.");

		pw.println( "<H2>Memory Bank Information</H2>");

		Enumeration bankEnum = owd.getMemoryBanks();
		if( !bankEnum.hasMoreElements())
		{
		    pw.println( "This device has no memory banks.");
		    return;
		}

		int bankIdx = -1;
		pw.println( "<OL>");
		while( bankEnum.hasMoreElements())
		{
		    MemoryBank mb = (MemoryBank)(bankEnum.nextElement());
		    pw.print( "<LI>");

			pw.print( "<A HREF=");
			pw.print( "OneWireServlet?");
			pw.print( buildQueryString(
				owd.getAdapter().getAdapterName(),
				owd.getAdapter().getPortName(),
				owd.getAddress(), ++bankIdx));
			pw.print( ">");
		    pw.print( mb.getBankDescription());
		    pw.print( "</A> (");
		    pw.print( mb.getSize());
		    pw.println( " bytes)");
		}
		pw.println( "</OL>");
	}

	public static final void sendBankDetail(
		PrintWriter pw, OneWireContainer owd, int bank)
	throws OneWireException
	{
		pw.print( "<H1>1-Wire Memory Bank Details for Device ");
		pw.print( owd.getAddressAsString());
		pw.println( "</H1>");

		pw.print( "This page displays details for memory bank ");
		pw.print( bank);
		pw.println( ".");

		pw.println( "<H2>Memory Bank Details</H2>");

		MemoryBank mb = null;
		int bankTmp = bank;
		Enumeration bankEnum = owd.getMemoryBanks();
		while( bankEnum.hasMoreElements() && bankTmp >= 0)
		{
		    mb = (MemoryBank)(bankEnum.nextElement());
		    --bankTmp;
		}

		if( bankTmp >= 0)
		{
		    pw.print( "This device has fewer than ");
		    pw.print( bank);
		    pw.println( " memory banks.");
		    return;
		}


		pw.print( "<TABLE><TR><TD>Bank Description:</TD><TD>");
		pw.print( mb.getBankDescription());
		pw.print( "</TD></TR><TR><TD>Size:</TD><TD>");
		pw.print( mb.getSize());
		pw.print( " bytes</TD></TR><TR><TD>Start Physical Address:</TD><TD>");
		pw.print( mb.getStartPhysicalAddress());
		pw.print( "</TD></TR><TR><TD>General Purpose Memory:</TD><TD>");
		pw.print( mb.isGeneralPurposeMemory());
		pw.print( "</TD></TR><TR><TD>Read/Write:</TD><TD>");
		pw.print( mb.isReadWrite());
		pw.print( "</TD></TR><TR><TD>Write Once:</TD><TD>");
		pw.print( mb.isWriteOnce());
		pw.print( "</TD></TR><TR><TD>Read Only:</TD><TD>");
		pw.print( mb.isReadOnly());
		pw.print( "</TD></TR><TR><TD>Non-Volatile:</TD><TD>");
		pw.print( mb.isNonVolatile());
		pw.print( "</TD></TR><TR><TD>Program Pulse Required:</TD><TD>");
		pw.print( mb.needsProgramPulse());
		pw.print( "</TD></TR><TR><TD>Power Delivery Required:</TD><TD>");
		pw.print( mb.needsPowerDelivery());

		if( mb instanceof PagedMemoryBank)
		{
		    PagedMemoryBank pmb = (PagedMemoryBank)mb;

			pw.print( "</TD></TR><TR><TD>Number of Pages:</TD><TD>");
			pw.print( pmb.getNumberPages());
			pw.print( "</TD></TR><TR><TD>Page Length:</TD><TD>");
			pw.print( pmb.getPageLength());
			pw.print( "</TD></TR><TR><TD>Max Packet Data Length:</TD><TD>");
			pw.print( pmb.getMaxPacketDataLength());
			pw.print( "</TD></TR><TR><TD>Page Auto CRC:</TD><TD>");
			pw.print( pmb.hasPageAutoCRC());
			pw.print( "</TD></TR><TR><TD>Extra Info:</TD><TD>");
		    if( pmb.hasExtraInfo())
		    {
		        pw.print( pmb.getExtraInfoDescription());
				pw.print( " (");
		        pw.print( pmb.getExtraInfoDescription());
				pw.print( " bytes)");
		    }
		    else
		    {
				pw.print( "None");
		    }
		}

		if( mb instanceof OTPMemoryBank)
		{
		    OTPMemoryBank otpmb = (OTPMemoryBank)mb;

			pw.print( "</TD></TR><TR><TD>Can Redirect Page:</TD><TD>");
			pw.print( otpmb.canRedirectPage());
			pw.print( "</TD></TR><TR><TD>Can Lock Page:</TD><TD>");
			pw.print( otpmb.canLockPage());
			pw.print( "</TD></TR><TR><TD>Can Lock Redirect Page:</TD><TD>");
			pw.print( otpmb.canLockRedirectPage());
		}

		pw.println( "</TD></TR></TABLE>");

		pw.println( "<H2>Memory Bank Contents</H2>");

		// Create buffer
		byte[] buf = new byte[mb.getSize()];
		// Read bytes
		mb.read( 0, false, buf, 0, buf.length);

		// Start table
		int cols = 32;
		if (buf.length < 32) cols = buf.length;
		pw.print( "<TABLE><TR><TH>Address</TH>");
		for( int i = 0; i < cols; ++i)
		{
		    pw.print( "<TH>");
		    pw.print( Integer.toHexString( i & 0xff));
		    pw.print( "</TH>");
		}

		for( int i = 0; i < buf.length; ++i)
		{
		    if( i % cols == 0)
		    {
				pw.print( "</TR><TR><TH ALIGN=RIGHT>");
				pw.print( Integer.toHexString(i));
				pw.print( "</TH>");
		    }
		    pw.print( "<TD>");
		    pw.print( Integer.toHexString( buf[i] & 0xff));
		    pw.print( "</TD>");
		}

		// End table
		pw.println( "</TR></TABLE>");
	}

	public static final String getRomIdHtml(
		String adapter, String port, byte[] id)
	{
		StringBuffer sb = new StringBuffer();

		boolean deviceSupported = isDeviceSupported( id);
		if( deviceSupported)
		{
			sb.append( "<A HREF=");
			sb.append( getDeviceServletName( id));
			sb.append( "?");
			sb.append( buildQueryString( adapter, port, id));
			sb.append( ">");
		}
		else
		{
			sb.append( "<B>");
		}
		sb.append( Address.toString( id));
		if( deviceSupported)
		{
			sb.append( "</A>");
		}
		else
		{
			sb.append( "</B>");
		}

		return sb.toString();
	}

	protected static final DSPortAdapter getAdapter( String adapterName)
	throws OneWireException
	{
		Enumeration paEnum = OneWireAccessProvider.enumerateAllAdapters();

		while( paEnum.hasMoreElements())
		{
			DSPortAdapter pa = (DSPortAdapter)paEnum.nextElement();
			String paName = pa.getAdapterName();
			if( paName != null && paName.equals( adapterName))
				return pa;
		}

		throw new OneWireException( "Adapter " +
			adapterName + " is not supported on this platform.");
	}

	protected static final DSPortAdapter getAdapter(
		String adapterName, String port)
	throws OneWireException
	{
		DSPortAdapter pa = getAdapter( adapterName);
		if( pa.selectPort( port))
		{
			if( pa.adapterDetected())
			{
				// Caller must free port
				return pa;
			}
		}

		throw new OneWireException( "Adapter " +
			adapterName + " is not detected on port " + port + ".");
	}

	protected void registerOneWireContainerClasses( DSPortAdapter pa)
	throws OneWireException
	{
		pa.registerOneWireContainerClass(
			0x20, com.smartsc.onewire.OneWireContainer20.class);
		pa.registerOneWireContainerClass(
			0xa0, com.smartsc.onewire.OneWireContainer20.class);
	}

	protected static final void print(
		PrintWriter pw, String name, String value)
	{
		pw.print( htmlEscape( name));
		pw.print( ": ");
		pw.print( htmlEscape( value));
		pw.print( "<BR>");
	}

	protected static final void print(
		PrintWriter pw, String name, boolean value)
	{
		print( pw, name, String.valueOf( value));
	}

	protected static final void sendAlert( PrintWriter pw, Throwable t)
	{
		sendAlert( pw, t.getMessage());
	}

	protected static final void sendAlert( PrintWriter pw, String s)
	{
		pw.println( "<SCRIPT>");
		pw.print  ( "alert('");
		pw.print  ( jsEscape( s));
		pw.println( "');");
		pw.println( "</SCRIPT>");
	}

	public static final String jsEscape( String s)
	{
		int len = s.length();
		StringBuffer sb = new StringBuffer( len);
		for( int i = 0; i < len; ++i)
		{
			char c = s.charAt( i);
			switch( c)
			{
				case '\b': sb.append( "\\b"); break;
				case '\f': sb.append( "\\f"); break;
				case '\n': sb.append( "\\n"); break;
				case '\r': sb.append( "\\r"); break;
				case '\t': sb.append( "\\t"); break;
				case '\'': sb.append( "\\'"); break;
				case '\"': sb.append( "\\\""); break;
				default  : sb.append( c);
			}
		}

		return sb.toString();
	}

	public static final String htmlEscape( String s)
	{
		int len = s.length();
		StringBuffer sb = new StringBuffer( len);
		StringTokenizer st = new StringTokenizer( s, "<>", true);
		while( st.hasMoreTokens())
		{
			String token = st.nextToken();
			if( token.equals( "<"))
				sb.append( "&lt;");
			else if( token.equals( ">"))
				sb.append( "&gt;");
			else
				sb.append( token);
		}
		return sb.toString();
	}

	protected static String buildQueryString(
		String adapter, String port, byte[] device)
	{
		return buildQueryString( adapter, port, device, -1);
	}

	protected static String buildQueryString(
		String adapter, String port, byte[] device, int bank)
	{
		StringBuffer sb = new StringBuffer( '?');
		buildQueryString( sb, ADAPTER, adapter);
		buildQueryString( sb, PORT,    port);
		if( device != null)
		{
			buildQueryString( sb, DEVICE, Address.toString( device));
		}
		if( bank != -1)
		{
			buildQueryString( sb, BANK, Integer.toString( bank));
		}
		return sb.toString();
	}

	protected static final String buildQueryString( String name, String value)
	{
		StringBuffer sb = new StringBuffer();
		return buildQueryString( sb, name, value);
	}

	protected static final String buildQueryString(
		StringBuffer sb, String name, String value)
	{
		if( name != null && value != null)
		{
			if( sb.length() > 2) sb.append( '&');
			sb.append( name);
			sb.append( '=');
			sb.append( value);
		}
		return sb.toString();
	}

	public static final String getFamilyCodeString( byte[] id)
	{
		StringBuffer sb = new StringBuffer( 2);
		sb.append( (char)('0' + ((id[0]&0xf0) >> 4) ) );
		sb.append( (char)('0' +  (id[0]&0x0f)       ) );
		return sb.toString();
	}

	public static final String getFamilyCodeServletName( byte[] id)
	{
		StringBuffer sb = new StringBuffer( "FamilyCode");
		sb.append( getFamilyCodeString( id));
		sb.append( "Servlet");
		return sb.toString();
	}

	public static final String getContainerClassName( byte[] id)
	{
		byte fc = id[0];
		id[0] &= 0x7f;
		StringBuffer sb = new StringBuffer(
		    "com.dalsemi.onewire.container.OneWireContainer");
		sb.append( getFamilyCodeString( id));
		id[0] = fc;
		return sb.toString();
	}

	protected static final boolean isFamilyCodeSupported( byte[] id)
	{
		try
		{
			Class.forName( getFamilyCodeServletName( id));
			return true;
		}
		catch( ClassNotFoundException cnfe)
		{
			return false;
		}
	}

	protected static final boolean isDeviceSupported( byte[] id)
	{
		if( isFamilyCodeSupported(id))
		{
		    return true;
		}

		try
		{
			Class.forName( getContainerClassName( id));
			return true;
		}
		catch( ClassNotFoundException cnfe)
		{
			return false;
		}
	}

	public static final String getDeviceServletName( byte[] id)
	{
		if( isFamilyCodeSupported(id))
		{
		    return getFamilyCodeServletName( id);
		}
		else
		{
		    return "OneWireServlet";
		}
	}

	public static final String ADAPTER = "adapter";
	public static final String PORT    = "port";
	public static final String DEVICE  = "device";
	public static final String BANK    = "bank";
}
